import tomli
import tomli_w
import yaml
import subprocess
from pathlib import Path
from rich.console import Console

from flowui.cli.utils.constants import COLOR_PALETTE, FLOWUI_HELM_PATH
from yaml.resolver import BaseResolver

class AsLiteral(str):
  pass

def represent_literal(dumper, data):
  return dumper.represent_scalar(BaseResolver.DEFAULT_SCALAR_TAG, data, style="|")

yaml.add_representer(AsLiteral, represent_literal)


console = Console()


def prepare_platform(cluster_name: str, github_workflows_token: str, github_operators_token: str) -> None:
    # Create local configuration file updated with user-provided arguments
    config_file_path = Path(__file__).resolve().parent / "config-flowui-local.toml"
    with open(str(config_file_path), "rb") as f:
        config_dict = tomli.load(f)
    
    running_path = str(Path().cwd().resolve())
    config_dict["path"]["FLOWUI_LOCAL_RUNNING_PATH"] = running_path
    config_dict["kind"]["FLOWUI_KIND_CLUSTER_NAME"] = cluster_name
    config_dict["github"]["FLOWUI_GITHUB_ACCESS_TOKEN_WORKFLOWS"] = github_workflows_token
    config_dict["github"]["FLOWUI_GITHUB_ACCESS_TOKEN_OPERATORS"] = github_operators_token
    
    with open("config-flowui-local.toml", "wb") as f:
        tomli_w.dump(config_dict, f)
    
    console.print("")
    console.print(f"FlowUI is prepared to run at: {running_path}")
    console.print(f"You can check and modify the configuration file at: {running_path}/config-flowui-local.toml")
    console.print("Next, run: `flowui platform create`")
    console.print("")


def create_platform(flowui_frontend_image: str = None, flowui_backend_image: str = None) -> None:
    # Load configuration values
    with open("config-flowui-local.toml", "rb") as f:
        flowui_config = tomli.load(f)

    # Create kind config file and run bash script to create Kind cluster
    kubeadm_config_patches = dict(
        kind="InitConfiguration",
        nodeRegistration=dict(
            kubeletExtraArgs={
                "node-labels": "ingress-ready=true"
            }
        )
    )
    kubeadm_parsed = AsLiteral(yaml.dump(kubeadm_config_patches))
    kind_config = dict(
        kind="Cluster",
        apiVersion="kind.x-k8s.io/v1alpha4",
        nodes=[
            dict(
                role="control-plane",
                kubeadmConfigPatches=[kubeadm_parsed],
                extraPortMappings=[
                    dict(
                        containerPort=80,
                        hostPort=80,
                        listenAddress="0.0.0.0",
                        protocol="TCP"
                    ),
                    dict(
                        containerPort=443,
                        hostPort=443,
                        listenAddress="0.0.0.0",
                        protocol="TCP"
                    )
                ]
            ),
            dict(
                role="worker",
                extraMounts=[
                    dict(
                        hostPath=flowui_config["path"]["FLOWUI_LOCAL_RUNNING_PATH"] + "/workflow_shared_storage",
                        containerPath="/cluster_shared_storage",
                        readOnly=False,
                        propagation="Bidirectional"
                    )
                ]
            ),
        ]
    )
    with open("kind-cluster-config.yaml", "w") as f:
        yaml.dump(kind_config, f)

    cluster_name = flowui_config["kind"]["FLOWUI_KIND_CLUSTER_NAME"]

    console.print("Removing previous Kind cluster...")
    subprocess.run(["kind", "delete", "cluster", "--name", cluster_name])
    console.print("Creating new Kind cluster...")
    subprocess.run(["kind", "create", "cluster", "--name", cluster_name, "--config", "kind-cluster-config.yaml"])
    console.print("")
    console.print("Kind cluster created successfully!", style=f"bold {COLOR_PALETTE.get('success')}")

    # TODO - Install services with helm upgrade
    console.print("")
    console.print("Installing NGINX controller...")
    subprocess.run(["kubectl", "apply", "-f", "https://raw.githubusercontent.com/kubernetes/ingress-nginx/main/deploy/static/provider/kind/deploy.yaml"], stdout=subprocess.DEVNULL)
    subprocess.run(["kubectl", "wait", "--namespace", "ingress-nginx", "--for", "condition=ready", "pod", "--selector=app.kubernetes.io/component=controller", "--timeout=90s"])

    # console.print("")
    # console.print("Installing Airflow services...")
    # subprocess.run([
    #     "helm", "upgrade", 
    #     "--set", "",
    #     "--set", "",
    #     "--set", "",
    #     "airflow", 
    #     "apache-airflow/airflow"
    # ])

    # console.print("")
    
    if flowui_frontend_image:
        console.print(f"Loading local frontend image {flowui_frontend_image} to Kind cluster...")
        subprocess.run(["kind", "load", "docker-image", flowui_frontend_image , "--name", cluster_name, "--nodes", f"{cluster_name}-worker"])
    else:
        flowui_frontend_image = "taufferconsulting/flowui-frontend:latest"
    
    if flowui_backend_image:
        console.print(f"Loading local backend image {flowui_backend_image} to Kind cluster...")
        subprocess.run(["kind", "load", "docker-image", flowui_backend_image , "--name", cluster_name, "--nodes", f"{cluster_name}-worker"])
    else:  
        flowui_backend_image = "taufferconsulting/flowui-backend:latest"

    # TODO separate operators and workflow tokens ?
    # TODO pass workflows repository as env variable
    token = flowui_config["github"]["FLOWUI_GITHUB_ACCESS_TOKEN_WORKFLOWS"]
    console.print("Installing FlowUI services...")
    subprocess.run([
        "helm", "install", 
        "--set", f"github_access_token=${token}",
        "--set", f"frontend.image={flowui_frontend_image}",
        "--set", f"backend.image={flowui_backend_image}",
        "flowui",
        FLOWUI_HELM_PATH
    ])

    console.print("")
    console.print("K8s resources created successfully!", style=f"bold {COLOR_PALETTE.get('success')}")

    # Forward ports of chosen services to localhost
    # if flowui_config["kind"]["FLOWUI_KIND_CLUSTER_EXPOSE_AIRFLOW_WEBSERVER"]:
    #     subprocess.run(["kubectl", "port-forward", "svc/airflow-webserver", "8080:8080"])

    console.print("You can now access the FlowUI frontend at: http://localhost/")
    console.print("and the Backend API at: http://localhost/api/")