from airflow import DAG
from kubernetes.client import models as k8s
from datetime import datetime
from typing import Callable
from flowui.client.flowui_backend_client import FlowuiBackendRestClient

from flowui.operators.python_operator import PythonOperator
from flowui.operators.bash_operator import BashOperator
from flowui.operators.k8s_operator import FlowuiKubernetesPodOperator
from flowui.logger import get_configured_logger


class Task(object):
    """
    The Task object represents a task in a workflow. 
    It is only instantiated by processes parsing dag files in Airflow.
    """

    def __init__(
        self, 
        dag: DAG, 
        task_id: str,
        operator: dict,
        operator_input_kwargs: dict,
        workflow_shared_storage: str = None,
        deploy_mode: str = 'local-k8s',
    ) -> None:
        # Task configuration and attributes
        self.task_id = task_id
        # Logger
        self.logger = get_configured_logger(f"{self.__class__.__name__ }-{self.task_id}")
        self.logger.info('### Configuring task object ###')
        self.dag = dag
        self.dag_id = self.dag.dag_id
        self.repository_id = operator["repository_id"]
        self.operator = operator
        self.workflow_shared_storage = workflow_shared_storage if workflow_shared_storage else ""
        self.deploy_mode = deploy_mode

        # Clients
        self.backend_client = FlowuiBackendRestClient(base_url="http://flowui-backend-service:8000/")

        # Set up Airflow operator using custom function
        self._task_operator = self._set_operator(operator_input_kwargs)

    def _set_operator(self, operator_input_kwargs) -> None:
        """
        Set airflow operator based on task configuration
        """

        if self.deploy_mode == "local-python":
            return PythonOperator(
                dag=self.dag,
                task_id=self.task_id,
                start_date=datetime(2021, 1, 1), # TODO - get correct start_date
                provide_context=True,
                op_kwargs=operator_input_kwargs,
                # queue=dependencies_group,
                make_python_callable_kwargs=dict(
                    operator_name=self.operator_name,
                    deploy_mode=self.deploy_mode,
                    task_id=self.task_id,
                    dag_id=self.dag_id,
                )
            )
        
        elif self.deploy_mode == "local-bash":
            cmds = 'source /opt/airflow/flowui_env/bin/activate && pip install -e /opt/flowui && flowui run-operator-bash'
            queue_name = self.operator.get("repository") + '-' + self.operator.get("dependency")
            return BashOperator(
                dag=self.dag,
                task_id=self.task_id,
                queue=queue_name,
                bash_command=cmds,
                env={
                    "FLOWUI_BASHOPERATOR_OPERATOR": str(self.operator),
                    "FLOWUI_BASHOPERATOR_INSTANTIATE_OP_KWARGS": str({
                        "deploy_mode": self.deploy_mode,
                        "task_id": self.task_id,
                        "dag_id": self.dag_id,
                    }),
                    "FLOWUI_BASHOPERATOR_RUN_OP_KWARGS": str(operator_input_kwargs),
                },
                append_env=True
            )

        # Ref: https://airflow.apache.org/docs/apache-airflow/1.10.14/_api/airflow/contrib/operators/kubernetes_pod_operator/index.html
        # Good example: https://github.com/apache/airflow/blob/main/tests/system/providers/cncf/kubernetes/example_kubernetes.py
        # Commands HAVE to go in a list object, ref: https://stackoverflow.com/a/55149915/11483674
        elif self.deploy_mode == "local-k8s":
            # https://airflow.apache.org/docs/apache-airflow/stable/templates-ref.html
            # https://www.astronomer.io/guides/templating/
            container_env_vars = {
                "FLOWUI_K8S_OPERATOR": self.operator["name"],
                "FLOWUI_K8S_INSTANTIATE_OP_KWARGS": str({
                    "deploy_mode": self.deploy_mode,
                    "task_id": self.task_id,
                    "dag_id": self.dag_id,
                }),
                "FLOWUI_K8S_RUN_OP_KWARGS": str(operator_input_kwargs),
                "FLOWUI_K8S_WORKFLOW_SHARED_STORAGE": self.workflow_shared_storage,
                "AIRFLOW_CONTEXT_EXECUTION_DATETIME": "{{ dag_run.logical_date | ts_nodash }}",
                "AIRFLOW_CONTEXT_DAG_RUN_ID": "{{ run_id }}",
            }

            all_volumes = []
            all_volume_mounts = []

            # Workflow shared volume local - shared by all tasks in a workflow
            if self.workflow_shared_storage == "local":
                volume_shared_storage = k8s.V1Volume(
                    name='workflow-shared-storage-volume',
                    #host_path=k8s.V1HostPathVolumeSource(path="/cluster_shared_storage"),  # path on host
                    persistent_volume_claim=k8s.V1PersistentVolumeClaimVolumeSource(claim_name='workflow-runs-volume-claim'),
                )
                volume_mount_shared_storage = k8s.V1VolumeMount(
                    name='workflow-shared-storage-volume', 
                    mount_path='/home/workflow_shared_storage',  # path inside container
                    read_only=False
                )
                all_volumes.append(volume_shared_storage)
                all_volume_mounts.append(volume_mount_shared_storage)

            ######################## For local Operators dev ###########################################
            # TODO condition test to check if operator is local or not
            # volume_dev_operators = k8s.V1Volume(
            #     name='jobs-persistent-storage-operators-dev',
            #     persistent_volume_claim=k8s.V1PersistentVolumeClaimVolumeSource(claim_name='workflow-operators-volume-claim'),
            # )
            # volume_mount_dev_operators = k8s.V1VolumeMount(
            #     name='jobs-persistent-storage-operators-dev', 
            #     mount_path='/home/flowui/operators_repository', 
            #     sub_path=None, 
            #     read_only=True
            # )
            # all_volumes.append(volume_dev_operators)
            # all_volume_mounts.append(volume_mount_dev_operators)
            ############################################################################################

            ######################## For local FlowUI dev ###############################################
            # volume_dev = k8s.V1Volume(
            #     name='jobs-persistent-storage-dev',
            #     persistent_volume_claim=k8s.V1PersistentVolumeClaimVolumeSource(claim_name='flowui-dev-volume-claim'),
            # )
            # volume_mount_dev = k8s.V1VolumeMount(
            #     name='jobs-persistent-storage-dev', 
            #     mount_path='/home/flowui_dev', 
            #     sub_path=None, 
            #     read_only=True
            # )
            # all_volumes.append(volume_dev)
            # all_volume_mounts.append(volume_mount_dev)
            ############################################################################################

            return FlowuiKubernetesPodOperator(
                operator_name=self.operator.get('name'),
                repository_id=self.repository_id,
                namespace='default',  # TODO - separate namespace by User or Workspace?
                image=self.operator["source_image"],
                image_pull_policy='IfNotPresent',
                task_id=self.task_id,
                name=f"airflow-worker-pod-{self.task_id}",
                # cmds=["python"],
                # arguments=["-c", "'import time; time.sleep(10000)'"],
                cmds=["flowui"],
                arguments=["run-operator-k8s"],
                env_vars=container_env_vars,
                do_xcom_push=True,
                in_cluster=True,
                # labels={"foo": "bar"},
                # secrets=[secret_file, secret_env, secret_all_keys],
                # ports=[port],
                volumes=all_volumes,
                volume_mounts=all_volume_mounts,
                # env_from=configmaps,
                # affinity=affinity,
                # is_delete_operator_pod=True,
                # hostnetwork=False,
                # tolerations=tolerations,
                # init_containers=[init_container],
                # priority_class_name="medium",
            )

    def __call__(self) -> Callable:
        return self._task_operator